from typing import (
    TYPE_CHECKING,
    Any,
    Self,
)
from google.cloud.firestore_v1.aggregation import AggregationQuery
from google.cloud.firestore_v1.query import Query as FirestoreQuery


if TYPE_CHECKING:
    from google.cloud.firestore_v1 import FieldFilter

    from model import Model


ASCENDING = 'ASCENDING'
DESCENDING = 'DESCENDING'


class Query(list):
    def __init__(
        self,
        records: list['Model'],
        model_class: type['Model'],
        query: FirestoreQuery = None,
    ) -> None:
        self._model_class = model_class
        self._query = query if query else self._model_class.collection.firestore_collection_ref
        super().__init__(records)

    def eval(self) -> Self:
        documents = self._query.stream()
        model_instances = []

        for document in documents:
            model_instance = self._model_class.from_db_dict(document.to_dict())
            model_instance._is_sync = True
            model_instances.append(model_instance)

        super().__init__(model_instances)

        return self

    def where(self, *filters: list['FieldFilter']) -> Self:
        for filter_ in filters:
            self._query = self._query.where(filter=filter_)

        self.eval()

        return self

    def order_by(self, field: str, desc: bool = False) -> Self:
        direction = DESCENDING if desc else ASCENDING

        self._query = self._query.order_by(field, direction=direction)
        self.eval()

        return self

    def limit(self, amount: int) -> Self:
        self._query = self._query.limit(amount)
        self.eval()

        return self

    def count(self) -> int:
        if isinstance(self._query, FirestoreQuery):
            aggregate_query = AggregationQuery(self._query)
        else:
            aggregate_query = self._query._aggregation_query()

        aggregate_query.count(alias='count')
        count_value = aggregate_query.get()[0][0].value

        return count_value

    def pluck(self, field_name: str) -> list[Any]:
        values = [getattr(record, field_name) for record in self]

        return values

    def delete(self) -> None:
        for record in self:
            record.delete()

    def __repr__(self) -> str:
        return f'Query({[str(record) for record in self]})'.replace("'", '')

    # Unused methods
    def append(self, *_) -> None:
        raise Exception("Results cannot be mutated manually.")

    def insert(self, *_) -> None:
        raise Exception("Results cannot be mutated manually.")

    def __add__(self, *_) -> None:
        raise Exception("Results cannot be mutated manually.")

    def __iadd__(self, *_) -> None:
        raise Exception("Results cannot be mutated manually.")



from google.cloud.firestore import Client as FirestoreClient

from .model import Model


class Firedom:
    def __init__(
        self,
        service_account_json_path: str | None = None,
        service_account_info: dict | None = None,
        firestore_client_instance: FirestoreClient | None = None,
    ) -> None:
        self.firestore_client: FirestoreClient = self.__get_firestore_client(
            service_account_json_path,
            service_account_info,
            firestore_client_instance,
        )
        self.Model: type[Model] = self.__get_model_class()

    def __get_firestore_client(
        self,
        service_account_json_path: str | None = None,
        service_account_info: dict | None = None,
        firestore_client_instance: FirestoreClient | None = None,
    ) -> FirestoreClient:
        firestore_client = None

        if firestore_client_instance:
            firestore_client = firestore_client_instance
        elif service_account_json_path:
            firestore_client = FirestoreClient.from_service_account_json(service_account_json_path)
        elif service_account_info:
            firestore_client = FirestoreClient.from_service_account_info(service_account_info)

        return firestore_client

    def __get_model_class(self) -> type[Model]:
        model_class = Model
        model_class._firestore_client = self.firestore_client

        return model_class



import unittest

from dataclasses import dataclass
from firedom import Firedom
from uuid import uuid4


firedom = Firedom(service_account_json_path='firebase.json')


@dataclass
class User(firedom.Model):
    username: str
    country: str
    number_of_pets: int = 0

    class Config:
        document_id_field = 'username'
        collection_id = f'users-{uuid4()}'


records_to_create = [
    {
        'username': 'user_1',
        'country': 'Chile',
    },
    {
        'username': 'user_2',
        'country': 'Argentina',
        'number_of_pets': 2,
    },
    {
        'username': 'user_3',
        'country': 'Bolivia',
    },
    {
        'username': 'user_4',
        'country': 'Chile',
        'number_of_pets': 3,
    },
]


class TestModel(unittest.TestCase):
    def setUp(self) -> None:
        for record_data in records_to_create:
            User(**record_data).save()

    def test_where_query(self) -> None:
        records = User.collection.where(User.username == 'user_1')
        assert len(records) == 1

        records = User.collection.where(User.username == 'non_existing_user')
        assert len(records) == 0

        records = User.collection.where(User.username != 'non_existing_user')
        assert len(records) == len(records_to_create)

        records = User.collection.where(User.country == 'Chile')
        assert len(records) == 2

        records = User.collection.where(User.number_of_pets >= 2)
        assert len(records) == 2

        records = User.collection.where(User.number_of_pets > 2)
        assert len(records) == 1

        records = User.collection.where(User.number_of_pets <= 2)
        assert len(records) == 3

        records = User.collection.where(User.number_of_pets < 2)
        assert len(records) == 2

        records = User.collection.where(User.country.is_in(['Chile', 'Argentina']))
        assert len(records) == 3

        records = User.collection.where(User.country == 'Chile', User.number_of_pets == 3)
        assert len(records) == 1

    def test_limit_query(self) -> None:
        records = User.collection.all().limit(2)
        assert len(records) == 2

    @classmethod
    def tearDownClass(cls) -> None:
        User.collection.all().delete()



import dataclasses
import types

from typing import (
    Any,
    get_origin,
    Self,
    TYPE_CHECKING,
)

from .collection import Collection
from .field import (
    Field,
    FieldFactory,
)


if TYPE_CHECKING:
    from google.cloud.firestore_v1 import (
        Client as FirestoreClient,
        DocumentReference,
    )


@dataclasses.dataclass(repr=False, kw_only=True)
class Model:
    _is_sync: bool = False
    _firestore_client: 'FirestoreClient' = None

    collection: Collection = None

    class Config:
        document_id_field: str = None
        collection_id: str | int | None = None

    def __init_subclass__(cls) -> None:
        current_repr = cls.__repr__

        dataclasses.dataclass(cls)

        cls.__repr__ = current_repr

        cls.__validate_document_id_field()

        fields_definitions = FieldFactory.generate_fields_definitions(cls)
        cls.collection = cls.__build_collection_instance(fields_definitions)
        cls.__set_fields(fields_definitions)

        super().__init_subclass__()

    def __new__(cls, **kwargs) -> type['Model']:
        cls.__validate_field_values(kwargs)

        return super().__new__(cls)

    @classmethod
    def __build_collection_instance(cls, fields_definition: dict[str, Field]) -> Collection:
        collection_id = cls.__name__.lower()

        if hasattr(cls.Config, 'collection_id') and cls.Config.collection_id:
            collection_id = cls.Config.collection_id

        collection = Collection(
            cls,
            fields_definition,
            cls.Config.document_id_field,
            collection_id,
        )

        return collection

    @classmethod
    def __set_fields(cls, fields_definition: dict[str, Field]) -> None:
        for name, field in fields_definition.items():
            setattr(cls, name, field)

    @classmethod
    def __validate_document_id_field(cls) -> None:
        if cls.Config.document_id_field not in cls.__dataclass_fields__.keys():
            raise ValueError(
                f"Document ID field \"{cls.Config.document_id_field}\" does not exist.",
            )
        document_id_field = cls.__dataclass_fields__[cls.Config.document_id_field]

        if document_id_field.type != str:
            raise TypeError(
                f"Document ID field value must be of type {str}. "
                f"Current type: {document_id_field.type}.",
            )

    @classmethod
    def __validate_field_values(cls, attributes: dict) -> None:
        for key, value in attributes.items():
            if key in cls.collection.fields.keys():
                expected_type = cls.collection.fields[key].field_type

                if isinstance(expected_type, types.GenericAlias):
                    expected_type = get_origin(expected_type)

                if expected_type != Any and type(value) is not expected_type:
                    raise TypeError(
                        f"Argument \"{key}\" must be of type {expected_type}. "
                        f"Current type: {type(value)}.",
                    )

    @property
    def document_id(self) -> str | int:
        return getattr(self, self.Config.document_id_field)

    @property
    def firestore_document_ref(self) -> 'DocumentReference':
        document_ref = self.__class__.collection.firestore_collection_ref.document(
            self.document_id,
        )

        return document_ref

    @classmethod
    def from_dict(cls, dict_: dict) -> Self:
        return cls(**dict_)

    @classmethod
    def from_db_dict(cls, dict_: dict) -> Self:
        fixed_dict = {
            key: cls.collection.fields[key].to_python_value(value)
            for key, value in dict_.items()
        }

        return cls(**fixed_dict)

    def to_dict(self) -> dict[str, Any]:
        registered_fields = self.__class__.collection.fields

        registered_fields_values = {}

        for name, field in registered_fields.items():
            if hasattr(self, name):
                fixed_value = field.to_db_value(getattr(self, name))
            else:
                fixed_value = getattr(self, name, field.default_value)

            registered_fields_values[name] = fixed_value

        return registered_fields_values

    def save(self) -> Self:
        self.firestore_document_ref.set(self.to_dict())
        self._is_sync = True

    def delete(self) -> None:
        if self._is_sync:
            self.firestore_document_ref.delete()

    def refresh_from_db(self) -> None:
        document = self.firestore_document_ref.get()
        self.from_db_dict(document.to_dict())
        self._is_sync = True

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}"
            f"({self.Config.document_id_field}={self.document_id})"
        )

    def __repr__(self) -> str:
        return str(self)



import unittest

from dataclasses import (
    dataclass,
    field,
)
from datetime import datetime
from firedom import Firedom
from uuid import uuid4


firedom = Firedom(service_account_json_path='firebase.json')


@dataclass
class User(firedom.Model):
    username: str
    country: str
    number_of_pets: int = 0
    actions: list = field(default_factory=list)
    extra_data: dict = field(default_factory=dict)
    created_at: datetime = datetime.now()

    class Config:
        document_id_field = 'username'
        collection_id = f'users-{uuid4()}'


records_to_create = [
    {
        'username': 'user_1',
        'country': 'Perú',
    },
    {
        'username': 'user_2',
        'country': 'Argentina',
        'number_of_pets': 2,
    },
    {
        'username': 'user_3',
        'country': 'Bolivia',
        'actions': ['a', 'b'],
    },
    {
        'username': 'user_4',
        'country': 'Chile',
        'number_of_pets': 3,
        'actions': ['a', 'b'],
        'extra_data': {'id': 1, 'is_active': True},
    },
]


class TestModel(unittest.TestCase):
    def setUp(self) -> None:
        for record_data in records_to_create:
            User.collection.create(**record_data)

    def test_record_creation(self) -> None:
        records = User.collection.all()
        assert len(records_to_create) == len(records)

    def test_record_update(self) -> None:
        first_record = User.collection.first()
        first_record.country = 'Japón'
        first_record.save()
        first_record.refresh_from_db()

        assert first_record.country == 'Japón'

    def test_record_delete(self) -> None:
        first_record = User.collection.first()
        record_to_delete_id = first_record.username
        first_record.delete()

        records = User.collection.all()

        assert len(records_to_create) - 1 == len(records)

        deleted_record = User.collection.get(record_to_delete_id)

        assert deleted_record is None

    @classmethod
    def tearDownClass(cls) -> None:
        User.collection.all().delete()



from typing import (
    Self,
    TYPE_CHECKING,
)

from .field import Field
from .query import Query


if TYPE_CHECKING:
    from google.cloud.firestore_v1 import (
        CollectionReference,
        FieldFilter,
    )

    from model import Model


class Collection:
    def __init__(
        self,
        model_class: type['Model'],
        fields: dict[str, Field],
        document_id_field: str,
        collection_id: str | None = None,
    ) -> None:
        self.model_class = model_class
        self.fields = fields
        self.document_id_field = document_id_field
        self.collection_id = (
            collection_id
            if collection_id
            else self.model_class.__name__.lower()
        )

    @property
    def firestore_collection_ref(self) -> 'CollectionReference':
        collection_ref = self.model_class._firestore_client.collection(self.collection_id)

        return collection_ref

    def get(self, document_id: str | int) -> Self | None:
        document_ref = self.firestore_collection_ref.document(document_id)
        document = document_ref.get()
        parsed_document = None

        if document.exists:
            parsed_document = self.model_class.from_db_dict(document.to_dict())
            parsed_document._is_sync = True

        return parsed_document

    def create(self, **kwargs) -> 'Model':
        document = self.model_class(**kwargs)
        document.save()

        return document

    def all(self) -> Query:
        query_result = Query([], self.model_class)
        query_result.eval()

        return query_result

    def where(self, *filters: list['FieldFilter']) -> Query:
        query_result = Query([], self.model_class)
        query_result.where(*filters)

        return query_result

    def first(self) -> 'Model':
        query_result = Query([], self.model_class)
        query_result.eval()

        if len(query_result):
            return query_result[0]

    def last(self) -> 'Model':
        query_result = Query([], self.model_class)
        query_result.eval()

        if len(query_result):
            return query_result[-1]



from setuptools import setup

setup(
    name='firedom',
    version='1.1.0',
    description='Simple Firestore ORM',
    url='https://github.com/afuenzalida/firedom',
    author='Andrés Fuenzalida',
    author_email='a.fuenzalida1494@gmail.com',
    license='MIT',
    packages=['firedom'],
    install_requires=[
        'google-cloud-firestore>=2.13.1',
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.11',
    ],
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
)



import dataclasses

from datetime import datetime
from google.api_core.datetime_helpers import DatetimeWithNanoseconds
from google.cloud.firestore_v1 import FieldFilter
from typing import (
    Any,
    TYPE_CHECKING,
)


if TYPE_CHECKING:
    from firedom.model import Model


class Field:
    def __init__(self, name: str, field_type: type, default_value: Any = None) -> None:
        self._name = name
        self.field_type = field_type
        self.default_value = default_value

    def __eq__(self, value: Any) -> FieldFilter:
        return FieldFilter(self._name, '==', value)

    def __ne__(self, value: Any) -> FieldFilter:
        return FieldFilter(self._name, '!=', value)

    def __lt__(self, value: Any) -> FieldFilter:
        return FieldFilter(self._name, '<', value)

    def __le__(self, value: Any) -> FieldFilter:
        return FieldFilter(self._name, '<=', value)

    def __gt__(self, value: Any) -> FieldFilter:
        return FieldFilter(self._name, '>', value)

    def __ge__(self, value: Any) -> FieldFilter:
        return FieldFilter(self._name, '>=', value)

    def __hash__(self) -> None:
        pass

    def is_in(self, values: list[Any]) -> FieldFilter:
        return FieldFilter(self._name, 'in', values)

    def is_not_in(self, values: list[Any]) -> FieldFilter:
        return FieldFilter(self._name, 'not-in', values)

    def to_db_value(self, value: Any) -> Any:
        return value

    def to_python_value(self, value: Any) -> Any:
        return value


class DatetimeField(Field):
    def to_python_value(self, value: DatetimeWithNanoseconds) -> datetime:
        return datetime.fromtimestamp(value.timestamp())


class RelatedField(Field):
    def to_db_value(self, value: 'Model') -> Any:
        return value.document_id

    def to_python_value(self, value: 'Model') -> Any:
        return self.field_type.collection.get(value)


class FieldFactory:
    FIELDS_MAPPING = {
        datetime: DatetimeField,
    }

    @classmethod
    def create_field(cls, **kwargs) -> Field:
        field_class = cls.FIELDS_MAPPING.get(kwargs.get('field_type'), Field)

        return field_class(**kwargs)

    @classmethod
    def generate_fields_definitions(cls, model_class: type['Model']) -> dict[str, Field]:
        fields_definitions = {}

        for name, field in model_class.__dataclass_fields__.items():
            if not field.kw_only:
                default_value = None

                if not isinstance(field.default, dataclasses._MISSING_TYPE):
                    default_value = field.default

                if any(parent_class.__name__ == 'Model' for parent_class in field.type.__bases__):
                    fields_definitions[name] = RelatedField(
                        name=name,
                        field_type=field.type,
                        default_value=default_value,
                    )
                else:
                    fields_definitions[name] = cls.create_field(
                        name=name,
                        field_type=field.type,
                        default_value=default_value,
                    )

        return fields_definitions



import unittest

from dataclasses import dataclass
from firedom import Firedom
from uuid import uuid4


firedom = Firedom(service_account_json_path='firebase.json')


@dataclass
class Country(firedom.Model):
    name: str

    class Config:
        document_id_field = 'name'
        collection_id = f'countries-{uuid4()}'


@dataclass
class City(firedom.Model):
    name: str
    country: Country

    class Config:
        document_id_field = 'name'
        collection_id = f'cities-{uuid4()}'


class TestModel(unittest.TestCase):
    def test_related_model_creation(self) -> None:
        created_country = Country.collection.create(name='Chile')
        created_city = City.collection.create(name='Santiago', country=created_country)

        assert created_city.country == created_country

        obtained_country = Country.collection.get('Chile')
        obtained_city = City.collection.get('Santiago')

        assert isinstance(obtained_city.country, obtained_country.__class__)

    @classmethod
    def tearDownClass(cls) -> None:
        City.collection.all().delete()
        Country.collection.all().delete()
